# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations


__all__ = ["ContextBool"]


class ContextBool:
    def __init__(self) -> None:
        self._value = False

    def __bool__(self) -> bool:
        return self._value

    def __enter__(self) -> None:
        self._value = True

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._value = False



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import typing as t

from .._async_compat.concurrency import AsyncCooperativeLock
from .._async_compat.util import AsyncUtil
from ..api import (
    AsyncBookmarkManager,
    Bookmarks,
)


TBmSupplier = t.Callable[[], t.Union[Bookmarks, t.Awaitable[Bookmarks]]]
TBmConsumer = t.Callable[[Bookmarks], t.Union[None, t.Awaitable[None]]]


def _bookmarks_to_set(
    bookmarks: Bookmarks | t.Iterable[str],
) -> set[str]:
    if isinstance(bookmarks, Bookmarks):
        return set(bookmarks.raw_values)
    return set(map(str, bookmarks))


class AsyncNeo4jBookmarkManager(AsyncBookmarkManager):
    def __init__(
        self,
        initial_bookmarks: Bookmarks | t.Iterable[str] | None = None,
        bookmarks_supplier: TBmSupplier | None = None,
        bookmarks_consumer: TBmConsumer | None = None,
    ) -> None:
        super().__init__()
        self._bookmarks_supplier = bookmarks_supplier
        self._bookmarks_consumer = bookmarks_consumer
        if not initial_bookmarks:
            self._bookmarks = set()
        else:
            if not hasattr(initial_bookmarks, "raw_values"):
                initial_bookmarks = Bookmarks.from_raw_values(
                    t.cast(t.Iterable[str], initial_bookmarks)
                )
            self._bookmarks = set(
                t.cast(Bookmarks, initial_bookmarks).raw_values
            )
        self._lock = AsyncCooperativeLock()

    async def update_bookmarks(
        self,
        previous_bookmarks: t.Collection[str],
        new_bookmarks: t.Collection[str],
    ) -> None:
        if not new_bookmarks:
            return
        with self._lock:
            self._bookmarks.difference_update(previous_bookmarks)
            self._bookmarks.update(new_bookmarks)
            if self._bookmarks_consumer:
                curr_bms_snapshot = Bookmarks.from_raw_values(self._bookmarks)
        if self._bookmarks_consumer:
            await AsyncUtil.callback(
                self._bookmarks_consumer, curr_bms_snapshot
            )

    async def get_bookmarks(self) -> set[str]:
        with self._lock:
            bms = set(self._bookmarks)
        if self._bookmarks_supplier:
            extra_bms = await AsyncUtil.callback(self._bookmarks_supplier)
            bms.update(extra_bms.raw_values)
        return bms



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Internal module for all internal exception classes."""

from os import strerror


class BoltError(Exception):
    """Base class for all Bolt protocol errors."""

    def __init__(self, message, address):
        super().__init__(message)
        self.address = address


class BoltConnectionError(BoltError):
    """Raised when a connection fails."""

    def __init__(self, message, address):
        msg = (
            "Connection Failed. "
            "Please ensure that your database is listening on the correct "
            "host and port and that you have enabled encryption if required. "
            "Note that the default encryption setting has changed in Neo4j "
            f"4.0. See the docs for more information. {message}"
        )

        super().__init__(msg, address)

    def __str__(self):
        s = super().__str__()
        errno = self.errno
        if errno:
            s += f" (code {errno}: {strerror(errno)})"
        return s

    @property
    def errno(self):
        try:
            return self.__cause__.errno
        except AttributeError:
            return None


class BoltSecurityError(BoltConnectionError):
    """Raised when a connection fails for security reasons."""

    def __str__(self):
        return f"[{self.__cause__.__class__.__name__}] {super().__str__()}"


class BoltHandshakeError(BoltError):
    """Raised when a handshake completes unsuccessfully."""

    def __init__(self, message, address, request_data, response_data):
        super().__init__(message, address)
        self.request_data = request_data
        self.response_data = response_data


class BoltProtocolError(BoltError):
    """Raised when an unexpected or unsupported protocol event occurs."""


class SocketDeadlineExceededError(RuntimeError):
    """Raised from sockets with deadlines when a timeout occurs."""



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import abc
import time
import typing as t
from dataclasses import dataclass

from .api import Auth
from .exceptions import AuthError


if t.TYPE_CHECKING:
    from os import PathLike

    from typing_extensions import Protocol as _Protocol

    from .api import _TAuth
    from .exceptions import Neo4jError
else:
    _Protocol = object


@dataclass
class ExpiringAuth:
    """
    Represents potentially expiring authentication information.

    This class is used with :meth:`.AuthManagers.bearer` and
    :meth:`.AsyncAuthManagers.bearer`.

    :param auth: The authentication information.
    :param expires_at:
        Unix timestamp (seconds since 1970-01-01 00:00:00 UTC)
        indicating when the authentication information expires.
        If :data:`None`, the authentication information is considered to not
        expire until the server explicitly indicates so.

    .. seealso::
        :meth:`.AuthManagers.bearer`,
        :meth:`.AsyncAuthManagers.bearer`

    .. versionadded:: 5.8

    .. versionchanged:: 5.9

        * Removed parameter and attribute ``expires_in`` (relative expiration
          time). Replaced with ``expires_at`` (absolute expiration time).
        * :meth:`.expires_in` can be used to create an :class:`.ExpiringAuth`
          with a relative expiration time.

    .. versionchanged:: 5.14 Stabilized from preview.
    """

    auth: _TAuth
    expires_at: float | None = None

    def expires_in(self, seconds: float) -> ExpiringAuth:
        """
        Return a (flat) copy of this object with a new expiration time.

        This is a convenience method for creating an :class:`.ExpiringAuth`
        for a relative expiration time ("expires in" instead of "expires at").

            >>> import time, freezegun
            >>> with freezegun.freeze_time("1970-01-01 00:00:40"):
            ...     ExpiringAuth(("user", "pass")).expires_in(2)
            ExpiringAuth(auth=('user', 'pass'), expires_at=42.0)
            >>> with freezegun.freeze_time("1970-01-01 00:00:40"):
            ...     ExpiringAuth(("user", "pass"), time.time() + 2)
            ExpiringAuth(auth=('user', 'pass'), expires_at=42.0)

        :param seconds:
            The number of seconds from now until the authentication information
            expires.

        .. versionadded:: 5.9
        """
        return ExpiringAuth(self.auth, time.time() + seconds)


def expiring_auth_has_expired(auth: ExpiringAuth) -> bool:
    expires_at = auth.expires_at
    return expires_at is not None and expires_at < time.time()


class AuthManager(metaclass=abc.ABCMeta):
    """
    Abstract base class for authentication information managers.

    The driver provides some default implementations of this class in
    :class:`.AuthManagers` for convenience.

    Custom implementations of this class can be used to provide more complex
    authentication refresh functionality.

    .. warning::

        The manager **must not** interact with the driver in any way as this
        can cause deadlocks and undefined behaviour.

        Furthermore, the manager is expected to be thread-safe.

        The token returned must always belong to the same identity.
        Switching identities using the `AuthManager` is undefined behavior.
        You may use :ref:`session-level authentication<session-auth-ref>`
        for such use-cases.

    .. seealso:: :class:`.AuthManagers`

    .. versionadded:: 5.8

    .. versionchanged:: 5.12
        ``on_auth_expired`` was removed from the interface and replaced by
        :meth:`handle_security_exception`. The new method is called when the
        server returns any `Neo.ClientError.Security.*` error. Its signature
        differs in that it additionally receives the error returned by the
        server and returns a boolean indicating whether the error was handled.

    .. versionchanged:: 5.14 Stabilized from preview.
    """

    @abc.abstractmethod
    def get_auth(self) -> _TAuth:
        """
        Return the current authentication information.

        The driver will call this method very frequently. It is recommended
        to implement some form of caching to avoid unnecessary overhead.

        .. warning::

            The method must only ever return auth information belonging to the
            same identity.
            Switching identities using the `AuthManager` is undefined behavior.
            You may use :ref:`session-level authentication<session-auth-ref>`
            for such use-cases.
        """
        ...

    @abc.abstractmethod
    def handle_security_exception(
        self, auth: _TAuth, error: Neo4jError
    ) -> bool:
        """
        Handle the server indicating authentication failure.

        The driver will call this method when the server returns any
        `Neo.ClientError.Security.*` error. The error will then be processed
        further as usual.

        :param auth:
            The authentication information that was used when the server
            returned the error.
        :param error:
            The error returned by the server.

        :returns:
            Whether the error was handled (:data:`True`), in which case the
            driver will mark the error as retryable
            (see :meth:`.Neo4jError.is_retryable`).

        .. versionadded:: 5.12
        """
        ...


class AsyncAuthManager(_Protocol, metaclass=abc.ABCMeta):
    """
    Async version of :class:`.AuthManager`.

    .. seealso:: :class:`.AuthManager`

    .. versionadded:: 5.8

    .. versionchanged:: 5.12
        ``on_auth_expired`` was removed from the interface and replaced by
         :meth:`handle_security_exception`. See :class:`.AuthManager`.

    .. versionchanged:: 5.14 Stabilized from preview.
    """

    @abc.abstractmethod
    async def get_auth(self) -> _TAuth:
        """
        Async version of :meth:`.AuthManager.get_auth`.

        .. seealso:: :meth:`.AuthManager.get_auth`
        """
        ...

    @abc.abstractmethod
    async def handle_security_exception(
        self, auth: _TAuth, error: Neo4jError
    ) -> bool:
        """
        Async version of :meth:`.AuthManager.handle_security_exception`.

        .. seealso:: :meth:`.AuthManager.handle_security_exception`
        """
        ...


@dataclass
class ClientCertificate:
    """
    Simple data class to hold client certificate information.

    The attributes are the same as the arguments to
    :meth:`ssl.SSLContext.load_cert_chain()`.

    .. versionadded:: 5.19

    .. versionchanged:: 5.27 Stabilized from preview.
    """

    certfile: str | bytes | PathLike[str] | PathLike[bytes]
    keyfile: str | bytes | PathLike[str] | PathLike[bytes] | None = None
    password: t.Callable[[], str | bytes] | str | bytes | None = None


class ClientCertificateProvider(_Protocol, metaclass=abc.ABCMeta):
    """
    Interface for providing a client certificate to the driver for mutual TLS.

    This is an abstract base class (:class:`abc.ABC`) as well as a protocol
    (:class:`typing.Protocol`). Meaning you can either inherit from it or just
    implement all required method on a class to satisfy the type constraints.

    The package provides some default implementations of this class in
    :class:`.ClientCertificateProviders` for convenience.

    The driver will call :meth:`.get_certificate` to check if the client wants
    the driver to use as new certificate for mutual TLS.

    The certificate is only used as a second factor for authenticating the
    client.
    The DBMS user still needs to authenticate with an authentication token.

    Note that the work done in the methods of this interface count towards the
    connection acquisition affected by the respective timeout setting
    :ref:`connection-acquisition-timeout-ref`.
    Should fetching the certificate be particularly slow, it might be necessary
    to increase the timeout.

    .. warning::

        The provider **must not** interact with the driver in any way as this
        can cause deadlocks and undefined behaviour.

    .. versionadded:: 5.19

    .. versionchanged:: 5.27 Stabilized from preview.
    """

    @abc.abstractmethod
    def get_certificate(self) -> ClientCertificate | None:
        """
        Return the new certificate (if present) to use for new connections.

        If no new certificate is available, return :data:`None`.
        This will make the driver continue using the current certificate.

        Note that a new certificate will only be used for new connections.
        Already established connections will continue using the old
        certificate as TLS is established during connection setup.

        :returns: The new certificate to use for new connections.
        """
        ...


class AsyncClientCertificateProvider(_Protocol, metaclass=abc.ABCMeta):
    """
    Async version of :class:`.ClientCertificateProvider`.

    The package provides some default implementations of this class in
    :class:`.AsyncClientCertificateProviders` for convenience.

    .. seealso::
        :class:`.ClientCertificateProvider`,
        :class:`.AsyncClientCertificateProviders`

    .. versionadded:: 5.19

    .. versionchanged:: 5.27 Stabilized from preview.
    """

    @abc.abstractmethod
    async def get_certificate(self) -> ClientCertificate | None:
        """
        Return the new certificate (if present) to use for new connections.

        .. seealso:: :meth:`.ClientCertificateProvider.get_certificate`
        """
        ...


def to_auth_dict(auth: _TAuth) -> dict[str, t.Any]:
    # Determine auth details
    if not auth:
        return {}
    elif isinstance(auth, tuple) and 2 <= len(auth) <= 3:
        return vars(Auth("basic", *auth))
    else:
        try:
            return vars(auth)
        except (KeyError, TypeError) as e:
            # TODO: 6.0 - change this to be a DriverError (or subclass)
            raise AuthError(
                f"Cannot determine auth details from {auth!r}"
            ) from e



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ruff: noqa: N818
# Not going to rename all Error classes that don't end on Error,
# which would break pretty much all users just to please the linter.


"""
Module containing the core driver exceptions.

Driver API Errors
=================
+ Neo4jError
  + ClientError
    + CypherSyntaxError
    + CypherTypeError
    + ConstraintError
    + AuthError
      + TokenExpired
    + Forbidden
  + DatabaseError
  + TransientError
    + DatabaseUnavailable
    + NotALeader
    + ForbiddenOnReadOnlyDatabase

+ DriverError
  + SessionError
  + TransactionError
    + TransactionNestingError
  + ResultError
    + ResultFailedError
    + ResultConsumedError
    + ResultNotSingleError
  + BrokenRecordError
  + SessionExpired
  + ServiceUnavailable
    + RoutingServiceUnavailable
    + WriteServiceUnavailable
    + ReadServiceUnavailable
    + IncompleteCommit
  + ConfigurationError
    + AuthConfigurationError
    + CertificateConfigurationError
"""

from __future__ import annotations

import typing as t
from copy import deepcopy as _deepcopy
from enum import Enum as _Enum

from ._meta import (
    deprecated,
    preview as _preview,
)


__all__ = [
    "AuthConfigurationError",
    "AuthError",
    "BrokenRecordError",
    "CertificateConfigurationError",
    "ClientError",
    "ConfigurationError",
    "ConstraintError",
    "CypherSyntaxError",
    "CypherTypeError",
    "DatabaseError",
    "DatabaseUnavailable",
    "DriverError",
    "Forbidden",
    "ForbiddenOnReadOnlyDatabase",
    "GqlError",
    "GqlErrorClassification",
    "IncompleteCommit",
    "Neo4jError",
    "NotALeader",
    "ReadServiceUnavailable",
    "ResultConsumedError",
    "ResultError",
    "ResultFailedError",
    "ResultNotSingleError",
    "RoutingServiceUnavailable",
    "ServiceUnavailable",
    "SessionError",
    "SessionExpired",
    "TokenExpired",
    "TransactionError",
    "TransactionNestingError",
    "TransientError",
    "UnsupportedServerProduct",
    "WriteServiceUnavailable",
]


if t.TYPE_CHECKING:
    from collections.abc import Mapping

    import typing_extensions as te

    from ._async.work import (
        AsyncManagedTransaction,
        AsyncResult,
        AsyncSession,
        AsyncTransaction,
    )
    from ._sync.work import (
        ManagedTransaction,
        Result,
        Session,
        Transaction,
    )

    _TTransaction = t.Union[
        AsyncManagedTransaction,
        AsyncTransaction,
        ManagedTransaction,
        Transaction,
    ]
    _TResult = t.Union[AsyncResult, Result]
    _TSession = t.Union[AsyncSession, Session]
    _T = t.TypeVar("_T")
else:
    _TTransaction = t.Union[
        "AsyncManagedTransaction",
        "AsyncTransaction",
        "ManagedTransaction",
        "Transaction",
    ]
    _TResult = t.Union["AsyncResult", "Result"]
    _TSession = t.Union["AsyncSession", "Session"]


__all__ = [
    "CLASSIFICATION_CLIENT",  # TODO: 6.0 - make constant private
    "CLASSIFICATION_DATABASE",  # TODO: 6.0 - make constant private
    "CLASSIFICATION_TRANSIENT",  # TODO: 6.0 - make constant private
    "ERROR_REWRITE_MAP",  # TODO: 6.0 - make constant private
    "AuthConfigurationError",
    "AuthError",
    "BrokenRecordError",
    "CertificateConfigurationError",
    "ClientError",
    "ConfigurationError",
    "ConstraintError",
    "CypherSyntaxError",
    "CypherTypeError",
    "DatabaseError",
    "DatabaseUnavailable",
    "DriverError",
    "Forbidden",
    "ForbiddenOnReadOnlyDatabase",
    "IncompleteCommit",
    "Neo4jError",
    "NotALeader",
    "ReadServiceUnavailable",
    "ResultConsumedError",
    "ResultError",
    "ResultFailedError",
    "ResultNotSingleError",
    "RoutingServiceUnavailable",
    "ServiceUnavailable",
    "SessionError",
    "SessionExpired",
    "TokenExpired",
    "TransactionError",
    "TransactionNestingError",
    "TransientError",
    "UnsupportedServerProduct",
    "WriteServiceUnavailable",
]


CLASSIFICATION_CLIENT: te.Final[str] = "ClientError"
CLASSIFICATION_TRANSIENT: te.Final[str] = "TransientError"
CLASSIFICATION_DATABASE: te.Final[str] = "DatabaseError"


ERROR_REWRITE_MAP: dict[str, tuple[str, str | None]] = {
    # This error can be retried ed. The driver just needs to re-authenticate
    # with the same credentials.
    "Neo.ClientError.Security.AuthorizationExpired": (
        CLASSIFICATION_TRANSIENT,
        None,
    ),
    # In 5.0, this error has been re-classified as ClientError.
    # For backwards compatibility with Neo4j 4.4 and earlier, we re-map it in
    # the driver, too.
    "Neo.TransientError.Transaction.Terminated": (
        CLASSIFICATION_CLIENT,
        "Neo.ClientError.Transaction.Terminated",
    ),
    # In 5.0, this error has been re-classified as ClientError.
    # For backwards compatibility with Neo4j 4.4 and earlier, we re-map it in
    # the driver, too.
    "Neo.TransientError.Transaction.LockClientStopped": (
        CLASSIFICATION_CLIENT,
        "Neo.ClientError.Transaction.LockClientStopped",
    ),
}


_UNKNOWN_NEO4J_CODE: te.Final[str] = "Neo.DatabaseError.General.UnknownError"
# TODO: 6.0 - Make _UNKNOWN_GQL_MESSAGE the default message
_UNKNOWN_MESSAGE: te.Final[str] = "An unknown error occurred"
_UNKNOWN_GQL_STATUS: te.Final[str] = "50N42"
_UNKNOWN_GQL_DESCRIPTION: te.Final[str] = (
    "error: general processing exception - unexpected error"
)
_UNKNOWN_GQL_MESSAGE: te.Final[str] = (
    f"{_UNKNOWN_GQL_STATUS}: "
    "Unexpected error has occurred. See debug log for details."
)
_UNKNOWN_GQL_DIAGNOSTIC_RECORD: te.Final[tuple[tuple[str, t.Any], ...]] = (
    ("OPERATION", ""),
    ("OPERATION_CODE", "0"),
    ("CURRENT_SCHEMA", "/"),
)


class GqlErrorClassification(str, _Enum):
    """
    Server-side GQL error category.

    Inherits from :class:`str` and :class:`enum.Enum`.
    Hence, can also be compared to its string value::

        >>> GqlErrorClassification.CLIENT_ERROR == "CLIENT_ERROR"
        True
        >>> GqlErrorClassification.DATABASE_ERROR == "DATABASE_ERROR"
        True
        >>> GqlErrorClassification.TRANSIENT_ERROR == "TRANSIENT_ERROR"
        True

    **This is a preview**.
    It might be changed without following the deprecation policy.
    See also
    https://github.com/neo4j/neo4j-python-driver/wiki/preview-features

    .. seealso:: :attr:`.GqlError.gql_classification`

    .. versionadded:: 5.26
    """

    CLIENT_ERROR = "CLIENT_ERROR"
    DATABASE_ERROR = "DATABASE_ERROR"
    TRANSIENT_ERROR = "TRANSIENT_ERROR"
    #: Used when the server provides a Classification which the driver is
    #: unaware of.
    #: This can happen when connecting to a server newer than the driver or
    #: before GQL errors were introduced.
    UNKNOWN = "UNKNOWN"


class GqlError(Exception):
    """
    The GQL compliant data of an error.

    This error isn't raised by the driver as is.
    Instead, only subclasses are raised.
    Further, it is used as the :attr:`__cause__` of GqlError subclasses.

    **This is a preview**.
    It might be changed without following the deprecation policy.
    See also
    https://github.com/neo4j/neo4j-python-driver/wiki/preview-features

    .. versionadded: 5.26
    """

    _gql_status: str
    # TODO: 6.0 - make message always str
    _message: str | None
    _gql_status_description: str
    _gql_raw_classification: str | None
    _gql_classification: GqlErrorClassification
    _status_diagnostic_record: dict[str, t.Any]  # original, internal only
    _diagnostic_record: dict[str, t.Any]  # copy to be used externally
    _gql_cause: GqlError | None

    @staticmethod
    def _hydrate_cause(**metadata: t.Any) -> GqlError:
        meta_extractor = _MetaExtractor(metadata)
        gql_status = meta_extractor.str_value("gql_status")
        description = meta_extractor.str_value("description")
        message = meta_extractor.str_value("message")
        diagnostic_record = meta_extractor.map_value("diagnostic_record")
        cause_map = meta_extractor.map_value("cause")
        if cause_map is not None:
            cause = GqlError._hydrate_cause(**cause_map)
        else:
            cause = None
        inst = GqlError()
        inst._init_gql(
            gql_status=gql_status,
            message=message,
            description=description,
            diagnostic_record=diagnostic_record,
            cause=cause,
        )
        return inst

    def _init_gql(
        self,
        *,
        gql_status: str | None = None,
        message: str | None = None,
        description: str | None = None,
        diagnostic_record: dict[str, t.Any] | None = None,
        cause: GqlError | None = None,
    ) -> None:
        if gql_status is None or message is None or description is None:
            self._gql_status = _UNKNOWN_GQL_STATUS
            self._message = _UNKNOWN_GQL_MESSAGE
            self._gql_status_description = _UNKNOWN_GQL_DESCRIPTION
        else:
            self._gql_status = gql_status
            self._message = message
            self._gql_status_description = description
        if diagnostic_record is not None:
            self._status_diagnostic_record = diagnostic_record
        self._gql_cause = cause

    def _set_unknown_gql(self):
        self._gql_status = _UNKNOWN_GQL_STATUS
        self._message = _UNKNOWN_GQL_MESSAGE
        self._gql_status_description = _UNKNOWN_GQL_DESCRIPTION

    def __getattribute__(self, item):
        if item != "__cause__":
            return super().__getattribute__(item)
        gql_cause = self._get_attr_or_none("_gql_cause")
        if gql_cause is None:
            # No GQL cause, no magic needed
            return super().__getattribute__(item)
        local_cause = self._get_attr_or_none("__cause__")
        if local_cause is None:
            # We have a GQL cause but no local cause
            # => set the GQL cause as the local cause
            self.__cause__ = gql_cause
            self.__suppress_context__ = True
            self._gql_cause = None
            return super().__getattribute__(item)
        # We have both a GQL cause and a local cause
        # => traverse the cause chain and append the local cause.
        root = gql_cause
        seen_errors = {id(self), id(root)}
        while True:
            cause = getattr(root, "__cause__", None)
            if cause is None:
                root.__cause__ = local_cause
                root.__suppress_context__ = True
                self.__cause__ = gql_cause
                self.__suppress_context__ = True
                self._gql_cause = None
                return gql_cause
            root = cause
            if id(root) in seen_errors:
                # Circular cause chain -> we have no choice but to either
                # overwrite the cause or ignore the new one.
                return local_cause
            seen_errors.add(id(root))

    def _get_attr_or_none(self, item):
        try:
            return super().__getattribute__(item)
        except AttributeError:
            return None

    @property
    def _gql_status_no_preview(self) -> str:
        if hasattr(self, "_gql_status"):
            return self._gql_status

        self._set_unknown_gql()
        return self._gql_status

    @property
    @_preview("GQLSTATUS support is a preview feature.")
    def gql_status(self) -> str:
        """
        The GQLSTATUS returned from the server.

        The status code ``50N42`` (unknown error) is a special code that the
        driver will use for polyfilling (when connected to an old,
        non-GQL-aware server).
        Further, it may be used by servers during the transition-phase to
        GQLSTATUS-awareness.

        .. note::
            This means that the code ``50N42`` is not guaranteed to be stable
            and may change in future versions of the driver or the server.
        """
        return self._gql_status_no_preview

    @property
    def _message_no_preview(self) -> str | None:
        if hasattr(self, "_message"):
            return self._message

        self._set_unknown_gql()
        return self._message

    @property
    @_preview("GQLSTATUS support is a preview feature.")
    def message(self) -> str | None:
        """
        The error message returned by the server.

        It is a string representation of the error that occurred.

        This message is meant for human consumption and debugging purposes.
        Don't rely on it in a programmatic way.

        This value is never :data:`None` unless the subclass in question
        states otherwise.
        """
        return self._message_no_preview

    @property
    def _gql_status_description_no_preview(self) -> str:
        if hasattr(self, "_gql_status_description"):
            return self._gql_status_description

        self._set_unknown_gql()
        return self._gql_status_description

    @property
    @_preview("GQLSTATUS support is a preview feature.")
    def gql_status_description(self) -> str:
        """
        A description of the GQLSTATUS returned from the server.

        It describes the error that occurred in detail.

        This description is meant for human consumption and debugging purposes.
        Don't rely on it in a programmatic way.
        """
        return self._gql_status_description_no_preview

    @property
    def _gql_raw_classification_no_preview(self) -> str | None:
        if hasattr(self, "_gql_raw_classification"):
            return self._gql_raw_classification

        diag_record = self._get_status_diagnostic_record()
        classification = diag_record.get("_classification")
        if not isinstance(classification, str):
            self._gql_raw_classification = None
        else:
            self._gql_raw_classification = classification
        return self._gql_raw_classification

    @property
    @_preview("GQLSTATUS support is a preview feature.")
    def gql_raw_classification(self) -> str | None:
        """
        Vendor specific classification of the error.

        This is a convenience accessor for ``_classification`` in the
        diagnostic record.
        :data:`None` is returned if the classification is not available
        or not a string.
        """
        return self._gql_raw_classification_no_preview

    @property
    def _gql_classification_no_preview(self) -> GqlErrorClassification:
        if hasattr(self, "_gql_classification"):
            return self._gql_classification

        classification = self._gql_raw_classification_no_preview
        if not (
            isinstance(classification, str)
            and classification
            in t.cast(t.Iterable[str], iter(GqlErrorClassification))
        ):
            self._gql_classification = GqlErrorClassification.UNKNOWN
        else:
            self._gql_classification = GqlErrorClassification(classification)
        return self._gql_classification

    @property
    @_preview("GQLSTATUS support is a preview feature.")
    def gql_classification(self) -> GqlErrorClassification:
        return self._gql_classification_no_preview

    def _get_status_diagnostic_record(self) -> dict[str, t.Any]:
        if hasattr(self, "_status_diagnostic_record"):
            return self._status_diagnostic_record

        self._status_diagnostic_record = dict(_UNKNOWN_GQL_DIAGNOSTIC_RECORD)
        return self._status_diagnostic_record

    @property
    def _diagnostic_record_no_preview(self) -> Mapping[str, t.Any]:
        if hasattr(self, "_diagnostic_record"):
            return self._diagnostic_record

        self._diagnostic_record = _deepcopy(
            self._get_status_diagnostic_record()
        )
        return self._diagnostic_record

    @property
    @_preview("GQLSTATUS support is a preview feature.")
    def diagnostic_record(self) -> Mapping[str, t.Any]:
        return self._diagnostic_record_no_preview

    def __str__(self):
        return (
            f"{{gql_status: {self._gql_status_no_preview}}} "
            f"{{gql_status_description: "
            f"{self._gql_status_description_no_preview}}} "
            f"{{message: {self._message_no_preview}}} "
            f"{{diagnostic_record: {self._diagnostic_record_no_preview}}} "
            f"{{raw_classification: "
            f"{self._gql_raw_classification_no_preview}}}"
        )


# Neo4jError
class Neo4jError(GqlError):
    """Raised when the Cypher engine returns an error to the client."""

    _neo4j_code: str | None
    _classification: str | None
    _category: str | None
    _title: str | None
    #: (dict) Any additional information returned by the server.
    _metadata: dict[str, t.Any] | None

    _retryable = False

    def __init__(self, *args) -> None:
        Exception.__init__(self, *args)
        self._neo4j_code = None
        self._classification = None
        self._category = None
        self._title = None
        self._metadata = None
        self._message = None

        # TODO: 6.0 - do this instead to get rid of all optional attributes
        # self._neo4j_code = _UNKNOWN_NEO4J_CODE
        # _, self._classification, self._category, self._title = (
        #     self._neo4j_code.split(".")
        # )
        # self._metadata = {}
        # self._init_gql()

    # TODO: 6.0 - Remove this alias
    @classmethod
    @deprecated(
        "Neo4jError.hydrate is deprecated and will be removed in a future "
        "version. It is an internal method and not meant for external use."
    )
    def hydrate(
        cls,
        code: str | None = None,
        message: str | None = None,
        **metadata: t.Any,
    ) -> Neo4jError:
        # backward compatibility: make falsy values None
        code = code or None
        message = message or None
        return cls._hydrate_neo4j(code=code, message=message, **metadata)

    @classmethod
    def _hydrate_neo4j(cls, **metadata: t.Any) -> Neo4jError:
        meta_extractor = _MetaExtractor(metadata)
        code = meta_extractor.str_value("code") or _UNKNOWN_NEO4J_CODE
        message = meta_extractor.str_value("message") or _UNKNOWN_MESSAGE
        inst = cls._basic_hydrate(
            neo4j_code=code,
            message=message,
        )
        inst._init_gql(
            gql_status=_UNKNOWN_GQL_STATUS,
            message=message,
            description=f"{_UNKNOWN_GQL_DESCRIPTION}. {message}",
        )
        inst._metadata = meta_extractor.rest()
        return inst

    @classmethod
    def _hydrate_gql(cls, **metadata: t.Any) -> Neo4jError:
        meta_extractor = _MetaExtractor(metadata)
        gql_status = meta_extractor.str_value("gql_status")
        status_description = meta_extractor.str_value("description")
        message = meta_extractor.str_value("message")
        if gql_status is None or status_description is None or message is None:
            gql_status = _UNKNOWN_GQL_STATUS
            # TODO: 6.0 - Make this fall back to _UNKNOWN_GQL_MESSAGE
            message = _UNKNOWN_MESSAGE
            status_description = _UNKNOWN_GQL_DESCRIPTION
        neo4j_code = meta_extractor.str_value(
            "neo4j_code",
            _UNKNOWN_NEO4J_CODE,
        )
        diagnostic_record = meta_extractor.map_value("diagnostic_record")
        cause_map = meta_extractor.map_value("cause")
        if cause_map is not None:
            cause = cls._hydrate_cause(**cause_map)
        else:
            cause = None

        inst = cls._basic_hydrate(
            neo4j_code=neo4j_code,
            message=message,
        )
        inst._init_gql(
            gql_status=gql_status,
            message=message,
            description=status_description,
            diagnostic_record=diagnostic_record,
            cause=cause,
        )
        inst._metadata = meta_extractor.rest()

        return inst

    @classmethod
    def _basic_hydrate(cls, *, neo4j_code: str, message: str) -> Neo4jError:
        try:
            _, classification, category, title = neo4j_code.split(".")
        except ValueError:
            classification = CLASSIFICATION_DATABASE
            category = "General"
            title = "UnknownError"
        else:
            classification_override, code_override = ERROR_REWRITE_MAP.get(
                neo4j_code, (None, None)
            )
            if classification_override is not None:
                classification = classification_override
            if code_override is not None:
                neo4j_code = code_override

        error_class: type[Neo4jError] = cls._extract_error_class(
            classification, neo4j_code
        )

        inst = error_class(message)
        inst._neo4j_code = neo4j_code
        inst._classification = classification
        inst._category = category
        inst._title = title
        inst._message = message

        return inst

    @classmethod
    def _extract_error_class(cls, classification, code) -> type[Neo4jError]:
        if classification == CLASSIFICATION_CLIENT:
            try:
                return client_errors[code]
            except KeyError:
                return ClientError

        elif classification == CLASSIFICATION_TRANSIENT:
            try:
                return transient_errors[code]
            except KeyError:
                return TransientError

        elif classification == CLASSIFICATION_DATABASE:
            return DatabaseError

        else:
            return cls

    @property
    def message(self) -> str | None:
        """
        The error message returned by the server.

        This value is only :data:`None` for locally created errors.
        """
        return self._message

    @message.setter
    @deprecated("Altering the message of a Neo4jError is deprecated.")
    def message(self, value: str) -> None:
        self._message = value

    @property
    def code(self) -> str | None:
        """
        The neo4j error code returned by the server.

        For example, "Neo.ClientError.Security.AuthorizationExpired".
        This value is only :data:`None` for locally created errors.
        """
        return self._neo4j_code

    # TODO: 6.0 - Remove this and all other deprecated setters
    @code.setter
    @deprecated("Altering the code of a Neo4jError is deprecated.")
    def code(self, value: str) -> None:
        self._neo4j_code = value

    @property
    def classification(self) -> str | None:
        # Undocumented, will likely be removed with support for neo4j codes
        return self._classification

    @classification.setter
    @deprecated("Altering the classification of Neo4jError is deprecated.")
    def classification(self, value: str) -> None:
        self._classification = value

    @property
    def category(self) -> str | None:
        # Undocumented, will likely be removed with support for neo4j codes
        return self._category

    @category.setter
    @deprecated("Altering the category of Neo4jError is deprecated.")
    def category(self, value: str) -> None:
        self._category = value

    @property
    def title(self) -> str | None:
        # Undocumented, will likely be removed with support for neo4j codes
        return self._title

    @title.setter
    @deprecated("Altering the title of Neo4jError is deprecated.")
    def title(self, value: str) -> None:
        self._title = value

    @property
    def metadata(self) -> dict[str, t.Any] | None:
        # Undocumented, might be useful for debugging
        return self._metadata

    @metadata.setter
    @deprecated("Altering the metadata of Neo4jError is deprecated.")
    def metadata(self, value: dict[str, t.Any]) -> None:
        self._metadata = value

    # TODO: 6.0 - Remove this alias
    @deprecated(
        "Neo4jError.is_retriable is deprecated and will be removed in a "
        "future version. Please use Neo4jError.is_retryable instead."
    )
    def is_retriable(self) -> bool:
        """
        Whether the error is retryable.

        See :meth:`.is_retryable`.

        :returns: :data:`True` if the error is retryable,
            :data:`False` otherwise.

        .. deprecated:: 5.0
            This method will be removed in a future version.
            Please use :meth:`.is_retryable` instead.
        """
        return self.is_retryable()

    def is_retryable(self) -> bool:
        """
        Whether the error is retryable.

        Indicates whether a transaction that yielded this error makes sense to
        retry. This method makes mostly sense when implementing a custom
        retry policy in conjunction with :ref:`explicit-transactions-ref`.

        .. warning::

            Auto-commit transactions
            (:meth:`.Session.run`/:meth:`.AsyncSession.run`) are not retryable
            regardless of this value.

        :returns: :data:`True` if the error is retryable,
            :data:`False` otherwise.

        .. versionadded:: 5.0
        """
        return self._retryable

    def _unauthenticates_all_connections(self) -> bool:
        return (
            self._neo4j_code == "Neo.ClientError.Security.AuthorizationExpired"
        )

    # TODO: 6.0 - Remove this alias
    invalidates_all_connections = deprecated(
        "Neo4jError.invalidates_all_connections is deprecated and will be "
        "removed in a future version. It is an internal method and not meant "
        "for external use."
    )(_unauthenticates_all_connections)

    def _is_fatal_during_discovery(self) -> bool:
        # checks if the code is an error that is caused by the client. In this
        # case the driver should fail fast during discovery.
        code = self._neo4j_code
        if not isinstance(code, str):
            return False
        if code in {
            "Neo.ClientError.Database.DatabaseNotFound",
            "Neo.ClientError.Transaction.InvalidBookmark",
            "Neo.ClientError.Transaction.InvalidBookmarkMixture",
            "Neo.ClientError.Statement.TypeError",
            "Neo.ClientError.Statement.ArgumentError",
            "Neo.ClientError.Request.Invalid",
        }:
            return True
        return (
            code.startswith("Neo.ClientError.Security.")
            and code != "Neo.ClientError.Security.AuthorizationExpired"
        )

    def _has_security_code(self) -> bool:
        if self._neo4j_code is None:
            return False
        return self._neo4j_code.startswith("Neo.ClientError.Security.")

    # TODO: 6.0 - Remove this alias
    is_fatal_during_discovery = deprecated(
        "Neo4jError.is_fatal_during_discovery is deprecated and will be "
        "removed in a future version. It is an internal method and not meant "
        "for external use."
    )(_is_fatal_during_discovery)

    def __str__(self):
        code = self._neo4j_code
        message = self._message
        if code or message:
            return f"{{code: {code}}} {{message: {message}}}"
            # TODO: 6.0 - Use gql status and status_description instead
            # something like:
            # return (
            #     f"{{gql_status: {self.gql_status}}} "
            #     f"{{neo4j_code: {self.neo4j_code}}} "
            #     f"{{gql_status_description: {self.gql_status_description}}} "
            #     f"{{diagnostic_record: {self.diagnostic_record}}}"
            # )
        return Exception.__str__(self)


class _MetaExtractor:
    def __init__(self, metadata: dict[str, t.Any]):
        self._metadata = metadata

    def rest(self) -> dict[str, t.Any]:
        return self._metadata

    @t.overload
    def str_value(self, key: str) -> str | None: ...

    @t.overload
    def str_value(self, key: str, default: _T) -> str | _T: ...

    def str_value(
        self, key: str, default: _T | None = None
    ) -> str | _T | None:
        res = self._metadata.pop(key, default)
        if not isinstance(res, str):
            res = default
        return res

    @t.overload
    def map_value(self, key: str) -> dict[str, t.Any] | None: ...

    @t.overload
    def map_value(self, key: str, default: _T) -> dict[str, t.Any] | _T: ...

    def map_value(
        self, key: str, default: _T | None = None
    ) -> dict[str, t.Any] | _T | None:
        res = self._metadata.pop(key, default)
        if not (
            isinstance(res, dict) and all(isinstance(k, str) for k in res)
        ):
            res = default
        return res


# Neo4jError > ClientError
class ClientError(Neo4jError):
    """
    Bad client request.

    The Client sent a bad request - changing the request might yield a
    successful outcome.
    """


# Neo4jError > ClientError > CypherSyntaxError
class CypherSyntaxError(ClientError):
    pass


# Neo4jError > ClientError > CypherTypeError
class CypherTypeError(ClientError):
    pass


# Neo4jError > ClientError > ConstraintError
class ConstraintError(ClientError):
    pass


# Neo4jError > ClientError > AuthError
class AuthError(ClientError):
    """Raised when authentication failure occurs."""


# Neo4jError > ClientError > AuthError > TokenExpired
class TokenExpired(AuthError):
    """Raised when the authentication token has expired."""


# Neo4jError > ClientError > Forbidden
class Forbidden(ClientError):
    pass


# Neo4jError > DatabaseError
class DatabaseError(Neo4jError):
    """The database failed to service the request."""


# Neo4jError > TransientError
class TransientError(Neo4jError):
    """
    Transient Error.

    The database cannot service the request right now, retrying later might
    yield a successful outcome.
    """

    _retryable = True


# Neo4jError > TransientError > DatabaseUnavailable
class DatabaseUnavailable(TransientError):
    pass


# Neo4jError > TransientError > NotALeader
class NotALeader(TransientError):
    pass


# Neo4jError > TransientError > ForbiddenOnReadOnlyDatabase
class ForbiddenOnReadOnlyDatabase(TransientError):
    pass


# TODO: 6.0 - Make map private
client_errors: dict[str, type[Neo4jError]] = {
    # ConstraintError
    "Neo.ClientError.Schema.ConstraintValidationFailed": ConstraintError,
    "Neo.ClientError.Schema.ConstraintViolation": ConstraintError,
    "Neo.ClientError.Statement.ConstraintVerificationFailed": ConstraintError,
    "Neo.ClientError.Statement.ConstraintViolation": ConstraintError,
    # CypherSyntaxError
    "Neo.ClientError.Statement.InvalidSyntax": CypherSyntaxError,
    "Neo.ClientError.Statement.SyntaxError": CypherSyntaxError,
    # CypherTypeError
    "Neo.ClientError.Procedure.TypeError": CypherTypeError,
    "Neo.ClientError.Statement.InvalidType": CypherTypeError,
    "Neo.ClientError.Statement.TypeError": CypherTypeError,
    # Forbidden
    "Neo.ClientError.General.ForbiddenOnReadOnlyDatabase": ForbiddenOnReadOnlyDatabase,  # noqa: E501
    "Neo.ClientError.General.ReadOnly": Forbidden,
    "Neo.ClientError.Schema.ForbiddenOnConstraintIndex": Forbidden,
    "Neo.ClientError.Schema.IndexBelongsToConstraint": Forbidden,
    "Neo.ClientError.Security.Forbidden": Forbidden,
    "Neo.ClientError.Transaction.ForbiddenDueToTransactionType": Forbidden,
    # AuthError
    "Neo.ClientError.Security.AuthorizationFailed": AuthError,
    "Neo.ClientError.Security.Unauthorized": AuthError,
    # TokenExpired
    "Neo.ClientError.Security.TokenExpired": TokenExpired,
    # NotALeader
    "Neo.ClientError.Cluster.NotALeader": NotALeader,
}

# TODO: 6.0 - Make map private
transient_errors: dict[str, type[Neo4jError]] = {
    # DatabaseUnavailableError
    "Neo.TransientError.General.DatabaseUnavailable": DatabaseUnavailable
}


# DriverError
class DriverError(GqlError):
    """Raised when the Driver raises an error."""

    def is_retryable(self) -> bool:
        """
        Whether the error is retryable.

        Indicates whether a transaction that yielded this error makes sense to
        retry. This method makes mostly sense when implementing a custom
        retry policy in conjunction with :ref:`explicit-transactions-ref`.

        .. warning::

            Auto-commit transactions
            (:meth:`.Session.run`/:meth:`.AsyncSession.run`) are not retryable
            regardless of this value.

        :returns: :data:`True` if the error is retryable,
            :data:`False` otherwise.

        .. versionadded:: 5.0
        """
        return False

    def __str__(self):
        return Exception.__str__(self)


# DriverError > SessionError
class SessionError(DriverError):
    """Raised when an error occurs while using a session."""

    session: _TSession

    def __init__(self, session_, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = session_


# DriverError > TransactionError
class TransactionError(DriverError):
    """Raised when an error occurs while using a transaction."""

    transaction: _TTransaction

    def __init__(self, transaction_, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.transaction = transaction_


# DriverError > TransactionError > TransactionNestingError
class TransactionNestingError(TransactionError):
    """Raised when transactions are nested incorrectly."""


# DriverError > ResultError
class ResultError(DriverError):
    """Raised when an error occurs while using a result object."""

    result: _TResult

    def __init__(self, result_, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.result = result_


# DriverError > ResultError > ResultFailedError
class ResultFailedError(ResultError):
    """
    Raised when trying to access records of a failed result.

    A :class:`.Result` will be considered failed if
     * itself encountered an error while fetching records
     * another result within the same transaction encountered an error while
       fetching records

    .. versionadded: 5.14
    """


# DriverError > ResultError > ResultConsumedError
class ResultConsumedError(ResultError):
    """Raised when trying to access records of a consumed result."""


# DriverError > ResultError > ResultNotSingleError
class ResultNotSingleError(ResultError):
    """Raised when a result should have exactly one record but does not."""


# DriverError > BrokenRecordError
class BrokenRecordError(DriverError):
    """
    Raised when accessing a Record's field that couldn't be decoded.

    This can for instance happen when the server sends a zoned datetime with a
    zone id unknown to the client.
    """


# DriverError > SessionExpired
class SessionExpired(DriverError):
    """
    The session has expired.

    Raised when a session is no longer able to fulfil the purpose described by
    its original parameters.
    """

    def __init__(self, *args):
        super().__init__(*args)
        self._init_gql(
            gql_status="08000",
            description="error: connection exception",
        )

    def is_retryable(self) -> bool:
        return True


# DriverError > ServiceUnavailable
class ServiceUnavailable(DriverError):
    """
    Raised when no database service is available.

    This may be due to incorrect configuration or could indicate a runtime
    failure of a database service that the driver is unable to route around.
    """

    def __init__(self, *args):
        super().__init__(*args)
        self._init_gql(
            gql_status="08000",
            description="error: connection exception",
        )

    def is_retryable(self) -> bool:
        return True


# DriverError > ServiceUnavailable > RoutingServiceUnavailable
class RoutingServiceUnavailable(ServiceUnavailable):
    """Raised when no routing service is available."""


# DriverError > ServiceUnavailable > WriteServiceUnavailable
class WriteServiceUnavailable(ServiceUnavailable):
    """Raised when no write service is available."""


# DriverError > ServiceUnavailable > ReadServiceUnavailable
class ReadServiceUnavailable(ServiceUnavailable):
    """Raised when no read service is available."""


# DriverError > ServiceUnavailable > IncompleteCommit
class IncompleteCommit(ServiceUnavailable):
    """
    Raised when the client looses connection while committing a transaction.

    Raised when a disconnection occurs while still waiting for a commit
    response. For non-idempotent write transactions, this leaves the data
    in an unknown state with regard to whether the transaction completed
    successfully or not.
    """

    def __init__(self, *args):
        super().__init__(*args)
        self._init_gql(
            gql_status="08007",
            description=(
                "error: connection exception - "
                "transaction resolution unknown"
            ),
        )

    def is_retryable(self) -> bool:
        return False


# DriverError > ConfigurationError
class ConfigurationError(DriverError):
    """Raised when there is an error concerning a configuration."""


# DriverError > ConfigurationError > AuthConfigurationError
class AuthConfigurationError(ConfigurationError):
    """Raised when there is an error with the authentication configuration."""


# DriverError > ConfigurationError > CertificateConfigurationError
class CertificateConfigurationError(ConfigurationError):
    """Raised when there is an error with the certificate configuration."""


class UnsupportedServerProduct(Exception):
    """Raised when an unsupported server product is detected."""



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import time
from ctypes import (
    byref,
    c_long,
    c_longlong,
    CDLL,
    Structure,
)
from platform import uname

from . import (
    Clock,
    ClockTime,
)
from ._arithmetic import nano_divmod


__all__ = [
    "LibCClock",
    "PEP564Clock",
    "SafeClock",
]


class SafeClock(Clock):
    """
    Clock implementation that should work for any variant of Python.

    This clock is guaranteed microsecond precision.
    """

    @classmethod
    def precision(cls):
        return 6

    @classmethod
    def available(cls):
        return True

    def utc_time(self):
        seconds, nanoseconds = nano_divmod(int(time.time() * 1000000), 1000000)
        return ClockTime(seconds, nanoseconds * 1000)


class PEP564Clock(Clock):
    """
    Clock implementation based on the PEP564 additions to Python 3.7.

    This clock is guaranteed nanosecond precision.
    """

    @classmethod
    def precision(cls):
        return 9

    @classmethod
    def available(cls):
        return hasattr(time, "time_ns")

    def utc_time(self):
        t = time.time_ns()
        seconds, nanoseconds = divmod(t, 1000000000)
        return ClockTime(seconds, nanoseconds)


class LibCClock(Clock):
    """
    Clock implementation backed by libc.

    Only works on platforms that provide libc.
    This clock is guaranteed nanosecond precision.
    """

    __libc = "libc.dylib" if uname()[0] == "Darwin" else "libc.so.6"

    class _TimeSpec(Structure):
        _fields_ = (
            ("seconds", c_longlong),
            ("nanoseconds", c_long),
        )

    @classmethod
    def precision(cls):
        return 9

    @classmethod
    def available(cls):
        try:
            _ = CDLL(cls.__libc)
        except OSError:
            return False
        else:
            return True

    def utc_time(self):
        libc = CDLL(self.__libc)
        ts = self._TimeSpec()
        status = libc.clock_gettime(0, byref(ts))
        if status == 0:
            return ClockTime(ts.seconds, ts.nanoseconds)
        else:
            raise RuntimeError(f"clock_gettime failed with status {status}")



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import asyncio
import dataclasses
import logging
import struct
import typing as t
from contextlib import suppress

from ... import addressing
from ..._async_compat.network import (
    AsyncBoltSocketBase,
    AsyncNetworkUtil,
)
from ..._exceptions import (
    BoltError,
    BoltProtocolError,
)
from ...exceptions import (
    DriverError,
    ServiceUnavailable,
)


if t.TYPE_CHECKING:
    from ssl import SSLContext

    import typing_extensions as te

    from ..._deadline import Deadline
    from ...addressing import (
        Address,
        ResolvedAddress,
    )


log = logging.getLogger("neo4j.io")


@dataclasses.dataclass
class HandshakeCtx:
    ctx: str
    deadline: Deadline
    local_port: int
    resolved_address: addressing.ResolvedAddress
    full_response: bytearray = dataclasses.field(default_factory=bytearray)


@dataclasses.dataclass
class BytesPrinter:
    bytes: bytes | bytearray

    def __str__(self):
        return f"0x{self.bytes.hex().upper()}"


class AsyncBoltSocket(AsyncBoltSocketBase):
    async def _parse_handshake_response_v1(
        self,
        ctx: HandshakeCtx,
        response: bytes,
    ) -> tuple[int, int]:
        agreed_version = response[-1], response[-2]
        log.debug(
            "[#%04X]  S: <HANDSHAKE> 0x%06X%02X",
            ctx.local_port,
            agreed_version[1],
            agreed_version[0],
        )
        return agreed_version

    async def _parse_handshake_response_v2(
        self,
        ctx: HandshakeCtx,
        response: bytes,
    ) -> tuple[int, int]:
        ctx.ctx = "handshake v2 offerings count"
        num_offerings = await self._read_varint(ctx)
        offerings = []
        for i in range(num_offerings):
            ctx.ctx = f"handshake v2 offering {i}"
            offering_response = await self._handshake_read(ctx, 4)
            offering = offering_response[-1:-4:-1]
            offerings.append(offering)
        ctx.ctx = "handshake v2 capabilities"
        _capabilities_offer = await self._read_varint(ctx)

        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug(
                "[#%04X]  S: <HANDSHAKE> %s [%i] %s %s",
                ctx.local_port,
                BytesPrinter(response),
                num_offerings,
                " ".join(
                    f"0x{vx[2]:04X}{vx[1]:02X}{vx[0]:02X}" for vx in offerings
                ),
                BytesPrinter(self._encode_varint(_capabilities_offer)),
            )

        supported_versions = sorted(self.Bolt.protocol_handlers.keys())
        chosen_version = 0, 0
        for v in supported_versions:
            for offer_major, offer_minor, offer_range in offerings:
                offer_max = (offer_major, offer_minor)
                offer_min = (offer_major, offer_minor - offer_range)
                if offer_min <= v <= offer_max:
                    chosen_version = v
                    break

        ctx.ctx = "handshake v2 chosen version"
        await self._handshake_send(
            ctx, bytes((0, 0, chosen_version[1], chosen_version[0]))
        )
        chosen_capabilities = 0
        capabilities = self._encode_varint(chosen_capabilities)
        ctx.ctx = "handshake v2 chosen capabilities"
        log.debug(
            "[#%04X]  C: <HANDSHAKE> 0x%06X%02X %s",
            ctx.local_port,
            chosen_version[1],
            chosen_version[0],
            BytesPrinter(capabilities),
        )
        await self._handshake_send(ctx, b"\x00")

        return chosen_version

    async def _read_varint(self, ctx: HandshakeCtx) -> int:
        next_byte = (await self._handshake_read(ctx, 1))[0]
        res = next_byte & 0x7F
        i = 0
        while next_byte & 0x80:
            i += 1
            next_byte = (await self._handshake_read(ctx, 1))[0]
            res += (next_byte & 0x7F) << (7 * i)
        return res

    @staticmethod
    def _encode_varint(n: int) -> bytearray:
        res = bytearray()
        while n >= 0x80:
            res.append(n & 0x7F | 0x80)
            n >>= 7
        res.append(n)
        return res

    async def _handshake_read(self, ctx: HandshakeCtx, n: int) -> bytes:
        original_timeout = self.gettimeout()
        self.settimeout(ctx.deadline.to_timeout())
        try:
            response = await self.recv(n)
            ctx.full_response.extend(response)
        except OSError as exc:
            raise ServiceUnavailable(
                f"Failed to read {ctx.ctx} from server "
                f"{ctx.resolved_address!r} (deadline {ctx.deadline})"
            ) from exc
        finally:
            self.settimeout(original_timeout)
        data_size = len(response)
        if data_size == 0:
            # If no data is returned after a successful select
            # response, the server has closed the connection
            log.debug("[#%04X]  S: <CLOSE>", ctx.local_port)
            await self.close()
            raise ServiceUnavailable(
                f"Connection to {ctx.resolved_address} closed with incomplete "
                f"handshake response"
            )
        if data_size != n:
            # Some garbled data has been received
            log.debug("[#%04X]  S: @*#!", ctx.local_port)
            await self.close()
            raise BoltProtocolError(
                f"Expected {ctx.ctx} from {ctx.resolved_address!r}, received "
                f"{response!r} instead (so far {ctx.full_response!r}); "
                "check for incorrect port number",
                address=ctx.resolved_address,
            )

        return response

    async def _handshake_send(self, ctx, data):
        original_timeout = self.gettimeout()
        self.settimeout(ctx.deadline.to_timeout())
        try:
            await self.sendall(data)
        except OSError as exc:
            raise ServiceUnavailable(
                f"Failed to write {ctx.ctx} to server "
                f"{ctx.resolved_address!r} (deadline {ctx.deadline})"
            ) from exc
        finally:
            self.settimeout(original_timeout)

    async def _handshake(
        self,
        resolved_address: ResolvedAddress,
        deadline: Deadline,
    ) -> tuple[tuple[int, int], bytes, bytes]:
        """
        Perform BOLT handshake.

        :param resolved_address:
        :param deadline: Deadline for handshake

        :returns: (version, client_handshake, server_response_data)
        """
        local_port = self.getsockname()[1]

        handshake = self.Bolt.get_handshake()
        if log.getEffectiveLevel() <= logging.DEBUG:
            handshake_bytes: t.Sequence = struct.unpack(">16B", handshake)
            handshake_bytes = [
                handshake[i : i + 4] for i in range(0, len(handshake_bytes), 4)
            ]

            supported_versions = [
                f"0x{vx[0]:02X}{vx[1]:02X}{vx[2]:02X}{vx[3]:02X}"
                for vx in handshake_bytes
            ]

            log.debug(
                "[#%04X]  C: <MAGIC> 0x%08X",
                local_port,
                int.from_bytes(self.Bolt.MAGIC_PREAMBLE, byteorder="big"),
            )
            log.debug(
                "[#%04X]  C: <HANDSHAKE> %s %s %s %s",
                local_port,
                *supported_versions,
            )

        request = self.Bolt.MAGIC_PREAMBLE + handshake

        ctx = HandshakeCtx(
            ctx="handshake opening",
            deadline=deadline,
            local_port=local_port,
            resolved_address=resolved_address,
        )

        await self._handshake_send(ctx, request)

        ctx.ctx = "four byte Bolt handshake response"
        response = await self._handshake_read(ctx, 4)

        if response == b"HTTP":
            log.debug("[#%04X]  C: <CLOSE> (received b'HTTP')", local_port)
            await self.close()
            raise ServiceUnavailable(
                f"Cannot to connect to Bolt service on {resolved_address!r} "
                "(looks like HTTP)"
            )
        elif response[-1] == 0xFF:
            # manifest style handshake
            manifest_version = response[-2]
            if manifest_version == 0x01:
                agreed_version = await self._parse_handshake_response_v2(
                    ctx,
                    response,
                )
            else:
                raise BoltProtocolError(
                    "Unsupported Bolt handshake manifest version "
                    f"{manifest_version} received from {resolved_address!r}.",
                    address=resolved_address,
                )
        else:
            agreed_version = await self._parse_handshake_response_v1(
                ctx,
                response,
            )

        return agreed_version, handshake, response

    @classmethod
    async def connect(
        cls,
        address: Address,
        *,
        tcp_timeout: float | None,
        deadline: Deadline,
        custom_resolver: t.Callable | None,
        ssl_context: SSLContext | None,
        keep_alive: bool,
    ) -> tuple[te.Self, tuple[int, int], bytes, bytes]:
        """
        Connect and perform a handshake.

        Return a valid Connection object, assuming a protocol version can be
        agreed.
        """
        errors = []
        failed_addresses = []
        # Establish a connection to the host and port specified
        # Catches refused connections see:
        # https://docs.python.org/2/library/errno.html

        resolved_addresses = AsyncNetworkUtil.resolve_address(
            addressing.Address(address), resolver=custom_resolver
        )
        async for resolved_address in resolved_addresses:
            deadline_timeout = deadline.to_timeout()
            if (
                deadline_timeout is not None
                and deadline_timeout <= tcp_timeout
            ):
                tcp_timeout = deadline_timeout
            s = None
            try:
                s = await cls._connect_secure(
                    resolved_address, tcp_timeout, keep_alive, ssl_context
                )
                agreed_version, handshake, response = await s._handshake(
                    resolved_address, deadline
                )
                return s, agreed_version, handshake, response
            except (BoltError, DriverError, OSError) as error:
                local_port = 0
                if isinstance(s, cls):
                    with suppress(OSError, AttributeError, TypeError):
                        local_port = s.getsockname()[1]
                err_str = error.__class__.__name__
                if str(error):
                    err_str += ": " + str(error)
                log.debug(
                    "[#%04X]  S: <CONNECTION FAILED> %s %s",
                    local_port,
                    resolved_address,
                    err_str,
                )
                if s:
                    await cls.close_socket(s)
                errors.append(error)
                failed_addresses.append(resolved_address)
            except asyncio.CancelledError:
                local_port = 0
                if isinstance(s, cls):
                    with suppress(OSError, AttributeError, TypeError):
                        local_port = s.getsockname()[1]
                log.debug(
                    "[#%04X]  C: <CANCELED> %s", local_port, resolved_address
                )
                if s:
                    with suppress(OSError):
                        s.kill()
                raise
            except Exception:
                if s:
                    await cls.close_socket(s)
                raise
        address_strs = tuple(map(str, failed_addresses))
        if not errors:
            raise ServiceUnavailable(
                f"Couldn't connect to {address} (resolved to {address_strs})"
            )
        else:
            error_strs = "\n".join(map(str, errors))
            raise ServiceUnavailable(
                f"Couldn't connect to {address} (resolved to {address_strs}):"
                f"\n{error_strs}"
            ) from errors[0]



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import inspect
import traceback
import typing as t
from copy import deepcopy
from functools import wraps

from ..._async_compat.concurrency import (
    Lock,
    RLock,
)
from ..._async_compat.util import Util
from ..._debug import ENABLED
from ..._meta import copy_signature


_TWrapped = t.TypeVar("_TWrapped", bound=t.Callable[..., t.Union[t.Any]])
_TWrappedIter = t.TypeVar(
    "_TWrappedIter", bound=t.Callable[..., t.Iterator]
)


class NonConcurrentMethodError(RuntimeError):
    pass


class NonConcurrentMethodChecker:
    if ENABLED:

        def __init__(self):
            self.__lock = RLock()
            self.__tracebacks_lock = Lock()
            self.__tracebacks = []

        def __make_error(self, tbs):
            msg = (
                f"Methods of {self.__class__} are not concurrency "
                "safe, but were invoked concurrently."
            )
            if tbs:
                msg += (
                    "\n\nOther invocation site:\n\n"
                    f"{''.join(traceback.format_list(tbs[0]))}"
                )
            return NonConcurrentMethodError(msg)

        @classmethod
        def _non_concurrent_method(cls, f: _TWrapped) -> _TWrapped:
            if Util.is_async_code:
                if not inspect.iscoroutinefunction(f):
                    raise TypeError(
                        "cannot decorate non-coroutine function with "
                        "NonConcurrentMethodChecked._non_concurrent_method"
                    )
            elif not callable(f):
                raise TypeError(
                    "cannot decorate non-callable object with "
                    "NonConcurrentMethodChecked._non_concurrent_method"
                )

            @copy_signature(f)
            @wraps(f)
            def inner(*args, **kwargs):
                self = args[0]
                assert isinstance(self, cls)

                with self.__tracebacks_lock:
                    acquired = self.__lock.acquire(blocking=False)
                    if acquired:
                        self.__tracebacks.append(Util.extract_stack())
                    else:
                        tbs = deepcopy(self.__tracebacks)
                if acquired:
                    try:
                        return f(*args, **kwargs)
                    finally:
                        with self.__tracebacks_lock:
                            self.__tracebacks.pop()
                            self.__lock.release()
                else:
                    raise self.__make_error(tbs)

            return inner

        @classmethod
        def _non_concurrent_iter(cls, f: _TWrappedIter) -> _TWrappedIter:
            if Util.is_async_code:
                if not inspect.isasyncgenfunction(f):
                    raise TypeError(
                        "cannot decorate non-async-generator function with "
                        "NonConcurrentMethodChecked._non_concurrent_iter"
                    )
            elif not inspect.isgeneratorfunction(f):
                raise TypeError(
                    "cannot decorate non-generator function with "
                    "NonConcurrentMethodChecked._non_concurrent_iter"
                )

            @copy_signature(f)
            @wraps(f)
            def inner(*args, **kwargs):
                self = args[0]
                assert isinstance(self, cls)

                iter_ = f(*args, **kwargs)
                while True:
                    with self.__tracebacks_lock:
                        acquired = self.__lock.acquire(blocking=False)
                        if acquired:
                            self.__tracebacks.append(Util.extract_stack())
                        else:
                            tbs = deepcopy(self.__tracebacks)
                    if acquired:
                        try:
                            item = iter_.__next__()
                        except StopIteration:
                            return
                        finally:
                            with self.__tracebacks_lock:
                                self.__tracebacks.pop()
                                self.__lock.release()
                        yield item
                    else:
                        raise self.__make_error(tbs)

            return inner

    else:

        @classmethod
        def _non_concurrent_method(cls, f: _TWrapped) -> _TWrapped:
            return f

        @classmethod
        def _non_concurrent_iter(cls, f: _TWrappedIter) -> _TWrappedIter:
            return f



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import typing as t

from ...._optional_deps import (
    np,
    pd,
)


NONE_VALUES: t.Tuple = (None,)
TRUE_VALUES: t.Tuple = (True,)
FALSE_VALUES: t.Tuple = (False,)
INT_TYPES: t.Tuple[t.Type, ...] = (int,)
FLOAT_TYPES: t.Tuple[t.Type, ...] = (float,)
# we can't put tuple here because spatial types subclass tuple,
# and we don't want to treat them as sequences
SEQUENCE_TYPES: t.Tuple[t.Type, ...] = (list,)
MAPPING_TYPES: t.Tuple[t.Type, ...] = (dict,)
BYTES_TYPES: t.Tuple[t.Type, ...] = (bytes, bytearray)


if np is not None:
    TRUE_VALUES = (*TRUE_VALUES, np.bool_(True))
    FALSE_VALUES = (*FALSE_VALUES, np.bool_(False))
    INT_TYPES = (*INT_TYPES, np.integer)
    FLOAT_TYPES = (*FLOAT_TYPES, np.floating)
    SEQUENCE_TYPES = (*SEQUENCE_TYPES, np.ndarray)

if pd is not None:
    NONE_VALUES = (*NONE_VALUES, pd.NA)
    SEQUENCE_TYPES = (
        *SEQUENCE_TYPES,
        pd.Series,
        pd.Categorical,
        pd.core.arrays.ExtensionArray,
    )
    MAPPING_TYPES = (*MAPPING_TYPES, pd.DataFrame)


__all__ = [
    "BYTES_TYPES",
    "FALSE_VALUES",
    "FLOAT_TYPES",
    "INT_TYPES",
    "MAPPING_TYPES",
    "NONE_VALUES",
    "SEQUENCE_TYPES",
    "TRUE_VALUES",
]



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import math
import typing as t
from time import monotonic

from .._async_compat.concurrency import CooperativeLock


if t.TYPE_CHECKING:
    import typing_extensions as te

    TKey: te.TypeAlias = t.Union[
        str,
        t.Tuple[t.Tuple[str, t.Hashable], ...],
        t.Tuple[None],
    ]
    TVal: te.TypeAlias = t.Tuple[float, str]


class HomeDbCache:
    _ttl: float
    _enabled: bool
    _max_size: int | None

    def __init__(
        self,
        enabled: bool = True,
        ttl: float = float("inf"),
        max_size: int | None = None,
    ) -> None:
        if math.isnan(ttl) or ttl <= 0:
            raise ValueError(f"home db cache ttl must be greater 0, got {ttl}")
        self._enabled = enabled
        self._ttl = ttl
        self._cache: dict[TKey, TVal] = {}
        self._lock = CooperativeLock()
        self._oldest_entry = monotonic()
        if max_size is not None and max_size <= 0:
            raise ValueError(
                f"home db cache max_size must be greater 0 or None, "
                f"got {max_size}"
            )
        self._max_size = max_size
        self._truncate_size = (
            min(max_size, int(0.01 * max_size * math.log(max_size)))
            if max_size is not None
            else None
        )

    def compute_key(
        self,
        imp_user: str | None,
        auth: dict | None,
    ) -> TKey:
        if not self._enabled:
            return (None,)
        if imp_user is not None:
            return imp_user
        if auth is not None:
            return _consolidate_auth_token(auth)
        return (None,)

    def get(self, key: TKey) -> str | None:
        if not self._enabled:
            return None
        with self._lock:
            self._clean(monotonic())
            val = self._cache.get(key)
            if val is None:
                return None
            return val[1]

    def set(self, key: TKey, value: str | None) -> None:
        if not self._enabled:
            return
        with self._lock:
            now = monotonic()
            self._clean(now)
            if value is None:
                self._cache.pop(key, None)
            else:
                self._cache[key] = (now, value)

    def clear(self) -> None:
        if not self._enabled:
            return
        with self._lock:
            self._cache = {}
            self._oldest_entry = monotonic()

    def _clean(self, now: float | None = None) -> None:
        now = monotonic() if now is None else now
        if now - self._oldest_entry > self._ttl:
            self._cache = {
                k: v
                for k, v in self._cache.items()
                if now - v[0] < self._ttl * 0.9
            }
            self._oldest_entry = min(
                (v[0] for v in self._cache.values()), default=now
            )
        if self._max_size and len(self._cache) > self._max_size:
            self._cache = dict(
                sorted(
                    self._cache.items(),
                    key=lambda item: item[1][0],
                    reverse=True,
                )[: self._truncate_size]
            )

    def __len__(self) -> int:
        return len(self._cache)

    @property
    def enabled(self) -> bool:
        return self._enabled


def _consolidate_auth_token(auth: dict) -> tuple | str:
    if auth.get("scheme") == "basic" and isinstance(
        auth.get("principal"), str
    ):
        return auth["principal"]
    return _hashable_dict(auth)


def _hashable_dict(d: dict) -> tuple:
    return tuple(
        (k, _hashable_dict(v) if isinstance(v, dict) else v)
        for k, v in sorted(d.items())
    )



#!/usr/bin/env python

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO: 6.0 - the whole deprecation and double naming shebang can be removed


import pathlib
import sys
from contextlib import contextmanager

import tomlkit
from setuptools import setup


THIS_DIR = pathlib.Path(__file__).parent

sys.path.insert(0, str(THIS_DIR / "src"))

from neo4j._meta import (  # noqa: E402 import has to happen after sys path manipulation
    deprecated_package as deprecated,
    package,
)


readme_path = THIS_DIR / "README.rst"
readme = readme_path.read_text(encoding="utf-8")

if deprecated:
    readme = (
        """\
.. warning::

    This package is deprecated and will stop receiving updates starting with
    version 6.0.0. Please install ``neo4j`` instead (which is an alias, i.e.,
    a drop-in replacement). See https://pypi.org/project/neo4j/ .

"""
        + readme
    )


pyproject_path = THIS_DIR / "pyproject.toml"


def change_project_name(new_name):
    with pyproject_path.open("a+", encoding="utf-8") as fd:
        fd.seek(0)
        pyproject = tomlkit.parse(fd.read())
        old_name = pyproject["project"]["name"]
        if old_name == new_name:
            return None
        pyproject["project"]["name"] = new_name
        fd.seek(0)
        fd.truncate()
        tomlkit.dump(pyproject, fd)
        return old_name


@contextmanager
def changed_package_name(new_name):
    old_name = change_project_name(new_name)
    try:
        yield
    finally:
        if old_name is not None:
            change_project_name(old_name)


with changed_package_name(package):
    setup(long_description=readme)



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import logging
import typing as t

from ..._async_compat.util import Util
from ..._auth_management import to_auth_dict
from ..._conf import WorkspaceConfig
from ..._meta import (
    deprecation_warn,
    unclosed_resource_warn,
)
from ...api import Bookmarks
from ...exceptions import (
    ServiceUnavailable,
    SessionError,
    SessionExpired,
)
from .._debug import NonConcurrentMethodChecker
from ..io import (
    AcquisitionAuth,
    AcquisitionDatabase,
)


if t.TYPE_CHECKING:
    from ...api import _TAuth
    from ...auth_management import AuthManager
    from ..home_db_cache import (
        HomeDbCache,
        TKey,
    )
else:
    _TAuth = t.Any


log = logging.getLogger("neo4j")


class Workspace(NonConcurrentMethodChecker):
    def __init__(self, pool, config):
        assert isinstance(config, WorkspaceConfig)
        self._pool = pool
        self._config = config
        self._connection = None
        self._connection_access_mode = None
        self._last_cache_key: TKey | None = None
        # Sessions are supposed to cache the database on which to operate.
        self._pinned_database = False
        self._bookmarks = ()
        self._initial_bookmarks = ()
        self._bookmark_manager = None
        self._last_from_bookmark_manager = None
        # Workspace has been closed.
        self._closed = False
        super().__init__()

    def __del__(
        self,
        _unclosed_resource_warn=unclosed_resource_warn,
        _is_async_code=Util.is_async_code,
        _deprecation_warn=deprecation_warn,
    ):
        if self._closed:
            return
        _unclosed_resource_warn(self)
        # TODO: 6.0 - remove this
        if _is_async_code:
            return
        try:
            _deprecation_warn(
                "Relying on Session's destructor to close the session "
                "is deprecated. Please make sure to close the session. Use it "
                "as a context (`with` statement) or make sure to call "
                "`.close()` explicitly. Future versions of the driver will "
                "not close sessions automatically."
            )
            self.close()
        except (OSError, ServiceUnavailable, SessionExpired):
            pass

    def __enter__(self) -> Workspace:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _make_db_resolution_callback(
        self,
    ) -> t.Callable[[str | None], None] | None:
        if self._pinned_database:
            return None

        def _database_callback(database: str | None) -> None:
            self._set_pinned_database(database)
            if self._last_cache_key is None or database is None:
                return
            db_cache: HomeDbCache = self._pool.home_db_cache
            db_cache.set(self._last_cache_key, database)

        return _database_callback

    def _set_pinned_database(self, database):
        if self._pinned_database:
            return
        log.debug("[#0000]  _: <WORKSPACE> pinning database: %r", database)
        self._pinned_database = True
        self._config.database = database

    def _initialize_bookmarks(self, bookmarks):
        if isinstance(bookmarks, Bookmarks):
            prepared_bookmarks = tuple(bookmarks.raw_values)
        elif hasattr(bookmarks, "__iter__"):
            deprecation_warn(
                "Passing an iterable as `bookmarks` to `Session` is "
                "deprecated. Please use a `Bookmarks` instance.",
                stack_level=5,
            )
            prepared_bookmarks = tuple(bookmarks)
        elif not bookmarks:
            prepared_bookmarks = ()
        else:
            raise TypeError(
                "Bookmarks must be an instance of Bookmarks or an "
                "iterable of raw bookmarks (deprecated)."
            )
        self._initial_bookmarks = self._bookmarks = prepared_bookmarks

    def _get_bookmarks(self):
        if self._bookmark_manager is None:
            return self._bookmarks

        bmm = Util.callback(self._bookmark_manager.get_bookmarks)
        initial = self._initial_bookmarks
        self._last_from_bookmark_manager = tuple({*bmm, *initial})
        return self._last_from_bookmark_manager

    def _update_bookmarks(self, new_bookmarks):
        if not new_bookmarks:
            return
        self._initial_bookmarks = ()
        self._bookmarks = new_bookmarks
        if self._bookmark_manager is None:
            return
        previous_bookmarks = self._last_from_bookmark_manager
        Util.callback(
            self._bookmark_manager.update_bookmarks,
            previous_bookmarks,
            new_bookmarks,
        )

    def _update_bookmark(self, bookmark):
        if not bookmark:
            return
        self._update_bookmarks((bookmark,))

    def _connect(self, access_mode, auth=None, **acquire_kwargs) -> None:
        acquisition_timeout = self._config.connection_acquisition_timeout
        force_auth = acquire_kwargs.pop("force_auth", False)
        acquire_auth = AcquisitionAuth(auth, force_auth=force_auth)

        if self._connection:
            # TODO: Investigate this
            # log.warning("FIXME: should always disconnect before connect")
            self._connection.send_all()
            self._connection.fetch_all()
            self._disconnect()

        ssr_enabled = self._pool.ssr_enabled
        target_db = self._get_routing_target_database(
            acquire_auth, ssr_enabled=ssr_enabled
        )
        acquire_kwargs_ = {
            "access_mode": access_mode,
            "timeout": acquisition_timeout,
            "database": target_db,
            "bookmarks": self._get_bookmarks(),
            "auth": acquire_auth,
            "liveness_check_timeout": None,
            "database_callback": self._make_db_resolution_callback(),
        }
        acquire_kwargs_.update(acquire_kwargs)
        self._connection = self._pool.acquire(**acquire_kwargs_)
        if (
            target_db.guessed
            and not self._pinned_database
            and not self._connection.ssr_enabled
        ):
            # race condition: we now have created a connection which does not
            # support SSR.
            # => we need to fall back to explicit home database resolution
            log.debug(
                "[#0000]  _: <WORKSPACE> detected ssr support race; "
                "falling back to explicit home database resolution",
            )
            self._disconnect()
            target_db = self._get_routing_target_database(
                acquire_auth, ssr_enabled=False
            )
            acquire_kwargs_["database"] = target_db
            self._connection = self._pool.acquire(**acquire_kwargs_)
        self._connection_access_mode = access_mode

    def _get_routing_target_database(
        self,
        acquire_auth: AcquisitionAuth,
        ssr_enabled: bool,
    ) -> AcquisitionDatabase:
        if (
            self._pinned_database
            or self._config.database is not None
            or self._pool.is_direct_pool
        ):
            log.debug(
                "[#0000]  _: <WORKSPACE> routing towards fixed database: %s",
                self._config.database,
            )
            self._set_pinned_database(self._config.database)
            return AcquisitionDatabase(self._config.database)

        auth = acquire_auth.auth
        resolved_auth = self._resolve_session_auth(auth)
        db_cache: HomeDbCache = self._pool.home_db_cache
        cache_key = db_cache.compute_key(
            self._config.impersonated_user,
            resolved_auth,
        )
        self._last_cache_key = cache_key

        if ssr_enabled:
            cached_db = db_cache.get(cache_key)
            if cached_db is not None:
                log.debug(
                    (
                        "[#0000]  _: <WORKSPACE> routing towards cached "
                        "database: %s"
                    ),
                    cached_db,
                )
                return AcquisitionDatabase(cached_db, guessed=True)

        acquisition_timeout = self._config.connection_acquisition_timeout
        log.debug("[#0000]  _: <WORKSPACE> resolve home database")
        self._pool.update_routing_table(
            database=self._config.database,
            imp_user=self._config.impersonated_user,
            bookmarks=self._get_bookmarks(),
            auth=acquire_auth,
            acquisition_timeout=acquisition_timeout,
            database_callback=self._make_db_resolution_callback(),
        )
        return AcquisitionDatabase(self._config.database)

    @staticmethod
    def _resolve_session_auth(
        auth: AuthManager | AuthManager | None,
    ) -> dict | None:
        if auth is None:
            return None
        # resolved_auth = await AsyncUtil.callback(auth.get_auth)
        # The above line breaks mypy
        # https://github.com/python/mypy/issues/15295
        auth_getter: t.Callable[[], _TAuth | t.Union[_TAuth]] = (
            auth.get_auth
        )
        # so we enforce the right type here
        # (explicit type annotation above added as it's a necessary assumption
        #  for this cast to be correct)
        resolved_auth = t.cast(_TAuth, Util.callback(auth_getter))
        return to_auth_dict(resolved_auth)

    def _disconnect(self, sync=False):
        self._last_cache_key = None
        if self._connection:
            if sync:
                try:
                    self._connection.send_all()
                    self._connection.fetch_all()
                except ServiceUnavailable:
                    pass
            if self._connection:
                self._pool.release(self._connection)
                self._connection = None
            self._connection_access_mode = None

    @NonConcurrentMethodChecker._non_concurrent_method
    def close(self) -> None:
        if self._closed:
            return
        self._disconnect(sync=True)
        self._closed = True

    def closed(self) -> bool:
        """
        Indicate whether the session has been closed.

        :returns: :data:`True` if closed, :data:`False` otherwise.
        """
        return self._closed

    def _check_state(self):
        if self._closed:
            raise SessionError(self, "Session closed")
